import base_network
import torch
import torch.nn as nn
import numpy as np
from eval import SAM_
from torch.nn import functional as F
import torch.utils.checkpoint as checkpoint
from timm.models.layers import DropPath, to_2tuple, trunc_normal_
import math
from einops import rearrange
from base_network import DWT_3D,IDWT_3D
 # import cv2 as cv

class GELU(nn.Module):
    def forward(self, x):
        return F.gelu(x)

# 残差 Swin Transforme 块 (RSTB)
class RSTB(nn.Module):
    """
    输入:
        dim (int): 输入特征的维度.
        input_resolution (tuple[int]): 输入分辨率.
        depth (int): SWin Transformer 块的个数.
        num_heads (int): 注意力头的个数.
        window_size (int): 本地(当前块中)窗口的大小.
        mlp_ratio (float): MLP隐藏层特征维度与嵌入层特征维度的比.
        qkv_bias (bool, optional): 给 query, key, value 添加一个可学习偏置，默认为 True.
        qk_scale (float | None, optional): 重写默认的缩放因子 scale.
        drop (float, optional): D 随机丢弃神经元，丢弃率默认为 0.0.
        attn_drop (float, optional): 注意力图随机丢弃率，默认为 0.0.
        drop_path (float | tuple[float], optional): 深度随机丢弃率，默认为 0.0.
        norm_layer (nn.Module, optional): 归一化操作，默认为 nn.LayerNorm.
        downsample (nn.Module | None, optional): 结尾处的下采样层，默认没有.
        use_checkpoint (bool): 是否使用 checkpointing 来节省显存，默认为 False.
        img_size: 输入图片的大小.
        patch_size: Patch 的大小.
        resi_connection: 残差连接之前的卷积块.
    """

    def __init__(self, dim, input_resolution, depth, num_heads, window_size,
                 mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., norm_layer=nn.LayerNorm, downsample=None, use_checkpoint=False,
                 img_size=224, patch_size=4, resi_connection='1conv'):
        super(RSTB, self).__init__()

        self.dim = dim  # 输入特征的维度
        self.input_resolution = input_resolution  # 输入分辨率

        # SWin Transformer 基础层
        self.residual_group = BasicLayer(dim=dim,
                                         input_resolution=input_resolution,
                                         depth=depth,
                                         num_heads=num_heads,
                                         window_size=window_size,
                                         mlp_ratio=mlp_ratio,
                                         qkv_bias=qkv_bias, qk_scale=qk_scale,
                                         drop=drop, attn_drop=attn_drop,
                                         drop_path=drop_path,
                                         norm_layer=norm_layer,
                                         downsample=downsample,
                                         use_checkpoint=use_checkpoint)

        if resi_connection == '1conv':  # 结尾用 1 个卷积层
            self.conv = nn.Conv2d(dim, dim, 3, 1, 1)
        elif resi_connection == '3conv':  # 结尾用 3 个卷积层
            # 为了减少参数使用和节约显存，采用瓶颈结构
            self.conv = nn.Sequential(nn.Conv2d(dim, dim // 4, 3, 1, 1), nn.LeakyReLU(negative_slope=0.2, inplace=True),
                                      nn.Conv2d(dim // 4, dim // 4, 1, 1, 0),
                                      nn.LeakyReLU(negative_slope=0.2, inplace=True),
                                      nn.Conv2d(dim // 4, dim, 3, 1, 1))

        # 图像转成 Patch Embeddings
        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=0, embed_dim=dim,
            norm_layer=None)

        # 从 Patch Embeddings 组合图像
        self.patch_unembed = PatchUnEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=0, embed_dim=dim,
            norm_layer=None)

    # 定义前向传播
    def forward(self, x, x_size):
        return self.patch_embed(self.conv(self.patch_unembed(self.residual_group(x, x_size), x_size))) + x  # 引入残差


# 单阶段的 SWin Transformer 基础层
class BasicLayer(nn.Module):
    """
    输入:
        dim (int): 输入特征的维度.
        input_resolution (tuple[int]): 输入分辨率.
        depth (int): SWin Transformer 块的个数.
        num_heads (int): 注意力头的个数.
        window_size (int): 本地(当前块中)窗口的大小.
        mlp_ratio (float): MLP隐藏层特征维度与嵌入层特征维度的比.
        qkv_bias (bool, optional): 给 query, key, value 添加一个可学习偏置，默认为 True.
        qk_scale (float | None, optional): 重写默认的缩放因子 scale.
        drop (float, optional): 随机丢弃神经元，丢弃率默认为 0.0.
        attn_drop (float, optional): 注意力图随机丢弃率，默认为 0.0.
        drop_path (float | tuple[float], optional): 深度随机丢弃率，默认为 0.0.
        norm_layer (nn.Module, optional): 归一化操作，默认为 nn.LayerNorm.
        downsample (nn.Module | None, optional): 结尾处的下采样层，默认没有.
        use_checkpoint (bool): 是否使用 checkpointing 来节省显存，默认为 False.
    """

    def __init__(self, dim, input_resolution, depth, num_heads, window_size,
                 mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., norm_layer=nn.LayerNorm, downsample=None, use_checkpoint=False):

        super().__init__()
        self.dim = dim  # 输入特征的维度
        self.input_resolution = input_resolution  # 输入分辨率
        self.depth = depth  # SWin Transformer 块的个数
        self.use_checkpoint = use_checkpoint  # 是否使用 checkpointing 来节省显存，默认为 False

        # 创建 Swin Transformer 网络
        self.blocks = nn.ModuleList([
            SwinTransformerBlock(dim=dim, input_resolution=input_resolution,
                                 num_heads=num_heads, window_size=window_size,
                                 shift_size=0 if (i % 2 == 0) else window_size // 2,
                                 mlp_ratio=mlp_ratio,
                                 qkv_bias=qkv_bias, qk_scale=qk_scale,
                                 drop=drop, attn_drop=attn_drop,
                                 drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path,
                                 norm_layer=norm_layer)
            for i in range(depth)])

        # patch 合并层
        if downsample is not None:  # 如果有下采样
            self.downsample = downsample(input_resolution, dim=dim, norm_layer=norm_layer)  # 下采样
        else:
            self.downsample = None  # 不做下采样

    #定义前向传播
    def forward(self, x, x_size):
        for blk in self.blocks:  # x 输入串联的 Swin Transformer 块
            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x, x_size)  # 使用 checkpoint
            else:
                x = blk(x, x_size)  # 直接输入网络
        if self.downsample is not None:
            x = self.downsample(x)  # 下采样
        return x

class Upsample(nn.Sequential):
    """
    输入:
        scale (int): 缩放因子，支持 2^n and 3.
        num_feat (int): 中间特征的通道数.
    """

    def __init__(self, scale, num_feat):
        m = []
        if (scale & (scale - 1)) == 0:  # 缩放因子等于 2^n
            for _ in range(int(math.log(scale, 2))):  #  循环 n 次
                m.append(nn.Conv2d(num_feat, 4 * num_feat, 3, 1, 1))  # 卷积层
                m.append(nn.PixelShuffle(2))  # pixelshuffle 上采样 2 倍
        elif scale == 3:  # 缩放因子等于 3
            m.append(nn.Conv2d(num_feat, 9 * num_feat, 3, 1, 1))  # 卷积层
            m.append(nn.PixelShuffle(3))  # pixelshuffle 上采样 3 倍
        else:
            # 报错，缩放因子不对
            raise ValueError(f'scale {scale} is not supported. ' 'Supported scales: 2^n and 3.')
        super(Upsample, self).__init__(*m)


# 将多个不重叠窗口重新合并
def window_reverse(windows, window_size, H, W):
    """
    输入:
        windows: (num_windows*B, window_size, window_size, C)  # 分割得到的窗口(已处理)
        window_size (int): Window size  # 窗口大小
        H (int): Height of image  # 原分割窗口前特征图的高
        W (int): Width of image  # 原分割窗口前特征图的宽
    返回:
        x: (B, H, W, C)  # 返回与分割前特征图结构一样的结果
    """
    # 以下就是分割窗口的逆向操作，不多解释
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.view(B, H // window_size, W // window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x

class block(nn.Sequential):
    def __init__(self, conv2d, wn, n_feats, kernel_size, padding, dilation, n_blocks):
        body = [
            base_network.ResBlock(
                conv2d, wn, n_feats, kernel_size, padding, dilation
            ) for _ in range(n_blocks)
        ]

        super(block, self).__init__(*body)


class left_right_group(nn.Module):
    def __init__(self, conv2d, wn, n_feats, kernel_size, n_blocks, padding, dilation):
        super(left_right_group, self).__init__()

        body = [
            block(
                conv2d, wn, n_feats, kernel_size, padding, dilation, n_blocks
            )
        ]

        self.body = nn.Sequential(*body)

    def forward(self, x):
        out = self.body(x) + x
        return out

class reduce_D(nn.Sequential):
    def __init__(self, conv2d, wn, input_feats, output_feats, kernel_size):
        body = [wn(conv2d(input_feats, output_feats, kernel_size))]

        super(reduce_D, self).__init__(*body)

class group(nn.Module):
    def __init__(self, conv2d, wn, n_feats, kernel_size, n_blocks):
        super(group, self).__init__()

        self.left_one_body = left_right_group(conv2d, wn, n_feats, kernel_size, n_blocks, padding=1, dilation=1)
        self.left_two_body = left_right_group(conv2d, wn, n_feats, kernel_size, n_blocks, padding=1, dilation=1)
        self.left_three_body = left_right_group(conv2d, wn, n_feats, kernel_size, n_blocks, padding=1, dilation=1)

        self.right_one_body = left_right_group(conv2d, wn, n_feats, kernel_size, n_blocks, padding=1, dilation=1)
        self.right_two_body = left_right_group(conv2d, wn, n_feats, kernel_size, n_blocks, padding=1, dilation=1)
        self.right_three_body = left_right_group(conv2d, wn, n_feats, kernel_size, n_blocks, padding=1, dilation=1)

        self.fusion_one = reduce_D(conv2d, wn, n_feats * 3, n_feats, kernel_size=1)
        self.fusion_two = reduce_D(conv2d, wn, n_feats * 3, n_feats, kernel_size=1)
        self.fusion_three = reduce_D(conv2d, wn, n_feats * 3, n_feats, kernel_size=1)

        self.gamma = nn.Parameter(torch.ones(3, 3))
        self.gamma_inter = nn.Parameter(torch.ones(3))

    def forward(self, group_one, group_two, group_three):
        intra_group_one = self.left_one_body(group_one)
        intra_group_two = self.left_two_body(group_two)
        intra_group_three = self.left_three_body(group_three)

        fusion_one = self.fusion_one(torch.cat([intra_group_one * self.gamma[0, 0], intra_group_two * self.gamma[0, 1],
                                                intra_group_three * self.gamma[0, 2]], 1))
        fusion_two = self.fusion_two(torch.cat([intra_group_one * self.gamma[1, 0], intra_group_two * self.gamma[1, 1],
                                                intra_group_three * self.gamma[1, 2]], 1))
        fusion_three = self.fusion_three(torch.cat(
            [intra_group_one * self.gamma[2, 0], intra_group_two * self.gamma[2, 1],
             intra_group_three * self.gamma[2, 2]], 1))

        inter_group_one = self.right_one_body(fusion_one)
        inter_group_two = self.right_two_body(fusion_two)
        inter_group_three = self.right_three_body(fusion_three)

        inter_group_one = torch.add(inter_group_one, group_one)
        inter_group_two = torch.add(inter_group_two, group_two)
        inter_group_three = torch.add(inter_group_three, group_three)

        group = torch.cat(
            [inter_group_one.unsqueeze(2) * self.gamma_inter[0], inter_group_two.unsqueeze(2) * self.gamma_inter[1],
             inter_group_three.unsqueeze(2) * self.gamma_inter[2]], 2)

        return group


class Unit(nn.Module):
    def __init__(self, conv3d, conv2d, wn, n_feats, kernel_size, act):
        super(Unit, self).__init__()

        self.threeunit = base_network.threeUnit(conv3d, wn, n_feats, act=act)

    def forward(self, x):
        out = self.threeunit(x) + x

        return out[:, :, 0, :, :], out[:, :, 1, :, :], out[:, :, 2, :, :]

class shallow_feature(nn.Module):
    def __init__(self, conv2d=base_network.default_conv2d, conv3d=base_network.default_conv3d):
        super(shallow_feature, self).__init__()

        n_feats = 64
        kernel_size = 3
        n_blocks = 2
        act = nn.ReLU(inplace=True)
        self.gamma_rnn = nn.Parameter(torch.ones(2))

        wn = lambda x: torch.nn.utils.weight_norm(x)

        # define head module for band group

        self.head_group_one = wn(conv2d(1, n_feats, kernel_size))
        # self.head_group_one = conv2d(1, n_feats, kernel_size)
        self.head_group_two = wn(conv2d(3, n_feats, kernel_size))
        self.head_group_three = wn(conv2d(3, n_feats, kernel_size))

        # define body module for  band group
        self.GFM_intra_one = group(conv2d, wn, n_feats, kernel_size, n_blocks)
        self.GFM_inter_one = Unit(conv3d, conv2d, wn, n_feats, kernel_size, act)
        self.conv2d = wn(conv2d(192,96,1))

    def forward(self,x,neigbor_l, neigbor_r):
        # x = x.unsqueeze(1)
        # q = torch.cat([x,x,x], 1)
        # k = torch.cat([neigbor_l[:, 1, :, :].unsqueeze(1), x, neigbor_r[:, 0, :, :].unsqueeze(1)], 1)
        # v = torch.cat([neigbor_l[:, 0, :, :].unsqueeze(1), x, neigbor_r[:, 1, :, :].unsqueeze(1)], 1)
        group_one = self.head_group_one(x)
        # wn(conv2d(1, n_feats, kernel_size))
        group_two = self.head_group_two(
            torch.cat([neigbor_l[:, 1, :, :].unsqueeze(1), x, neigbor_r[:, 0, :, :].unsqueeze(1)], 1))
        group_three = self.head_group_three(
            torch.cat([neigbor_l[:, 0, :, :].unsqueeze(1), x, neigbor_r[:, 1, :, :].unsqueeze(1)], 1))

        res = []
        res.append(group_one)
        res.append(group_two)
        res.append(group_three)

        ## feature extraction for fusion
        out = self.GFM_intra_one(group_one, group_two, group_three)
        group_one, group_two, group_three = self.GFM_inter_one(out)
        group_one = torch.add(group_one, res[0])
        group_two = torch.add(group_two, res[1])
        group_three = torch.add(group_three, res[2])
        out = torch.cat([group_one,group_two,group_three],1)
        out = self.conv2d(out)
        return out

class MS_MSA(nn.Module):
    def __init__(
            self,
            dim,
            dim_head,
            heads,
    ):
        super().__init__()
        self.num_heads = heads
        self.dim_head = dim_head
        self.to_q = nn.Linear(dim, dim_head * heads, bias=False)
        self.to_k = nn.Linear(dim, dim_head * heads, bias=False)
        self.to_v = nn.Linear(dim, dim_head * heads, bias=False)
        self.rescale = nn.Parameter(torch.ones(heads, 1, 1))
        self.proj = nn.Linear(dim_head * heads, dim, bias=True)
        self.pos_emb = nn.Sequential(
            nn.Conv2d(dim, dim, 3, 1, 1, bias=False, groups=dim),
            GELU(),
            nn.Conv2d(dim, dim, 3, 1, 1, bias=False, groups=dim),
        )
        self.dim = dim

    def forward(self, x_in):
        """
        x_in: [b,h,w,c]
        return out: [b,h,w,c]
        """
        b, h, w, c = x_in.shape
        x = x_in.reshape(b,h*w,c)
        q_inp = self.to_q(x)
        k_inp = self.to_k(x)
        v_inp = self.to_v(x)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.num_heads),
                                (q_inp, k_inp, v_inp))
        v = v
        # q: b,heads,hw,c
        q = q.transpose(-2, -1)
        k = k.transpose(-2, -1)
        v = v.transpose(-2, -1)
        q = F.normalize(q, dim=-1, p=2)
        k = F.normalize(k, dim=-1, p=2)
        attn = (k @ q.transpose(-2, -1))   # A = K^T*Q
        attn = attn * self.rescale
        attn = attn.softmax(dim=-1)
        x = attn @ v   # b,heads,d,hw
        x = x.permute(0, 3, 1, 2)    # Transpose
        x = x.reshape(b, h * w, self.num_heads * self.dim_head)
        out_c = self.proj(x).view(b, h, w, c)
        out_p = self.pos_emb(v_inp.reshape(b,h,w,c).permute(0, 3, 1, 2)).permute(0, 2, 3, 1)
        out = out_c + out_p

        return out

# Swin Transformer 块
class SwinTransformerBlock(nn.Module):
    """
    输入:
        dim (int): 输入特征的维度.
        input_resolution (tuple[int]): 输入特征图的分辨率.
        num_heads (int): 注意力头的个数.
        window_size (int): 窗口的大小.
        shift_size (int): SW-MSA 的移位值.
        mlp_ratio (float): 多层感知机隐藏层的维度和嵌入层的比.
        qkv_bias (bool, optional): 给 query, key, value 添加一个可学习偏置，默认为 True.
        qk_scale (float | None, optional): 重写默认的缩放因子 scale.
        drop (float, optional): 随机神经元丢弃率，默认为 0.0.
        attn_drop (float, optional): 注意力图随机丢弃率，默认为 0.0.
        drop_path (float, optional): 深度随机丢弃率，默认为 0.0.
        act_layer (nn.Module, optional): 激活函数，默认为 nn.GELU.
        norm_layer (nn.Module, optional): 归一化操作，默认为 nn.LayerNorm.
    """

    def __init__(self, dim, input_resolution, num_heads, window_size=7, shift_size=0,
                 mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0., drop_path=0.,
                 act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.dim = dim  # 输入特征的维度
        self.input_resolution = input_resolution  # 输入特征图的分辨率
        self.num_heads = num_heads  # 注意力头的个数
        self.window_size = window_size  # 窗口的大小
        self.shift_size = shift_size  # SW-MSA 的移位大小
        self.mlp_ratio = mlp_ratio  # 多层感知机隐藏层的维度和嵌入层的比
        if min(self.input_resolution) <= self.window_size:  # 如果输入分辨率小于等于窗口大小
            self.shift_size = 0  # 移位大小为 0
            self.window_size = min(self.input_resolution)  # 窗口大小等于输入分辨率大小
        # 断言移位值必须小于等于窗口的大小
        assert 0 <= self.shift_size < self.window_size, "shift_size must in 0-window_size"

        ###################  STL层
        self.norm1 = norm_layer(dim)  # 归一化层
        # 窗口注意力
        self.attn = WindowAttention(
            dim, window_size=to_2tuple(self.window_size), num_heads=num_heads,
            qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)

        # 如果丢弃率大于 0 则进行随机丢弃，否则进行占位(不做任何操作)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)  # 归一化层
        mlp_hidden_dim = int(dim * mlp_ratio)  # 多层感知机隐藏层维度
        # 多层感知机
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)
        #############################
        self.norm3 = norm_layer(dim)
        self.attn1 = MS_MSA(dim = dim, dim_head=dim // num_heads, heads=num_heads )
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm4 = norm_layer(dim)  # 归一化层
        mlp_hidden_dim = int(dim * mlp_ratio)  # 多层感知机隐藏层维度
        # 多层感知机
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if self.shift_size > 0:  # 如果移位值大于 0
            attn_mask = self.calculate_mask(self.input_resolution)  # 计算注意力 mask
        else:
            attn_mask = None  # 注意力 mask 赋空

        self.register_buffer("attn_mask", attn_mask)  # 保存注意力 mask，不参与更新

    # 计算注意力 mask
    def calculate_mask(self, x_size):
        H, W = x_size  # 特征图的高宽
        img_mask = torch.zeros((1, H, W, 1))  # 新建张量，结构为 [1, H, W, 1]
        # 以下两 slices 中的数据是索引，具体缘由尚未搞懂
        h_slices = (slice(0, -self.window_size),  # 索引 0 到索引倒数第 window_size
                    slice(-self.window_size, -self.shift_size),  # 索引倒数第 window_size 到索引倒数第 shift_size
                    slice(-self.shift_size, None))  # 索引倒数第 shift_size 后所有索引
        w_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        cnt = 0
        for h in h_slices:
            for w in w_slices:
                img_mask[:, h, w, :] = cnt  # 将 img_mask 中 h, w 对应索引范围的值置为 cnt
                cnt += 1  # 加 1

        mask_windows = window_partition(img_mask, self.window_size)  # 窗口分割，返回值结构为 [nW, window_size, window_size, 1]
        mask_windows = mask_windows.view(-1,
                                         self.window_size * self.window_size)  # 重构结构为二维张量，列数为 [window_size*window_size]
        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)  # 增加第 2 维度减去增加第 3 维度的注意力 mask
        # 用浮点数 -100. 填充注意力 mask 中值不为 0 的元素，再用浮点数 0. 填充注意力 mask 中值为 0 的元素
        attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))

        return attn_mask

    # 定义前向传播
    def forward(self, x, x_size):
        H, W = x_size  # 输入特征图的分辨率
        B, L, C = x.shape  # 输入特征的 batch 个数，长度和维度
        # assert L == H * W, "input feature has wrong size"

        shortcut = x
        x = self.norm1(x)  # 归一化
        x = x.view(B, H, W, C)  # 重构 x 为结构 [B, H, W, C]

        # 循环移位
        if self.shift_size > 0:  # 如果移位值大于 0
            # 第 0 维度上移 shift_size 位，第 1 维度左移 shift_size 位
            shifted_x = torch.roll(x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2))
        else:
            shifted_x = x  # 不移位

        # 对移位操作得到的特征图分割窗口, nW 是窗口的个数
        x_windows = window_partition(shifted_x, self.window_size)  # 结构为 [nW*B, window_size, window_size, C]
        x_windows = x_windows.view(-1, self.window_size * self.window_size, C)  # 结构为 [nW*B, window_size*window_size, C]

        # W-MSA/SW-MSA, 用在分辨率是窗口大小的整数倍的图像上进行测试
        if self.input_resolution == x_size:  # 输入分辨率与设定一致，不需要重新计算注意力 mask
            attn_windows = self.attn(x_windows, mask=self.attn_mask)  # 注意力窗口，结构为 [nW*B, window_size*window_size, C]
        else:  # 输入分辨率与设定不一致，需要重新计算注意力 mask
            attn_windows = self.attn(x_windows, mask=self.calculate_mask(x_size).to(x.device))

        # 合并窗口
        attn_windows = attn_windows.view(-1, self.window_size, self.window_size,
                                         C)  # 结构为 [-1, window_size, window_size, C]
        shifted_x = window_reverse(attn_windows, self.window_size, H, W)  # 结构为 [B, H', W', C]

        # 逆向循环移位
        if self.shift_size > 0:
            # 第 0 维度下移 shift_size 位，第 1 维度右移 shift_size 位
            x = torch.roll(shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2))
        else:
            x = shifted_x  # 不逆向移位
        x = x.view(B, H * W, C)  # 结构为 [B, H*W， C]

        # FFN
        x = shortcut + self.drop_path(x)  # 对 x 做 dropout，引入残差
        x = x + self.drop_path(self.mlp(self.norm2(x)))  # 归一化后通过 MLP，再做 dropout，引入残差
        x = self.norm3(x)
        x = x.view(B, H, W, C)
        x = self.attn1(x)
        x = x.view(B, H * W, C)
        x = x + self.drop_path(self.mlp(self.norm4(x)))


        return x


class PatchUnEmbed(nn.Module):
    def __init__(self, img_size=32, patch_size=4, in_chans=5, embed_dim=96, norm_layer=None):
        super().__init__()
        patches_resolution = [img_size // patch_size, img_size // patch_size]
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_channel = in_chans
        self.embed_dim = embed_dim

    def forward(self,x,x_size):
        B , HW , C = x.shape
        x = x.transpose(1,2).view(B, self.embed_dim, x_size[0] , x_size[1])
        return x

def window_partition(x, window_size):
    B, H, W, C = x.shape
    x = x.view(B, H//window_size, window_size, W // window_size, window_size, C)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
    return windows

class WindowAttention(nn.Module):
    def __init__(self,dim, window_size, num_heads, qkv_bias = True, qk_scale = None, attn_drop = 0., proj_drop = 0.):
        super().__init__()
        self.dim = dim
        self.window_size = window_size
        self.num_heads = num_heads
        head_dim = dim // num_heads  # 注意力头的维度
        self.scale = qk_scale or head_dim ** -0.5  # 缩放因子 scale

        # 定义相对位置偏移的参数表，结构为 [2*Wh-1 * 2*Ww-1, num_heads]
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1), num_heads))

        # 获取窗口内每个 token 的成对的相对位置索引
        coords_h = torch.arange(self.window_size[0])  # 高维度上的坐标 (0, 7)
        coords_w = torch.arange(self.window_size[1])  # 宽维度上的坐标 (0, 7)
        coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 坐标，结构为 [2, Wh, Ww]
        coords_flatten = torch.flatten(coords, 1)  # 重构张量结构为 [2, Wh*Ww]
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 相对坐标，结构为 [2, Wh*Ww, Wh*Ww]
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # 交换维度，结构为 [Wh*Ww, Wh*Ww, 2]
        relative_coords[:, :, 0] += self.window_size[0] - 1  # 第1个维度移位
        relative_coords[:, :, 1] += self.window_size[1] - 1  # 第1个维度移位
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1  # 第1个维度的值乘以 2倍的 Ww，再减 1
        relative_position_index = relative_coords.sum(-1)  # 相对位置索引，结构为 [Wh*Ww, Wh*Ww]
        self.register_buffer("relative_position_index", relative_position_index)  # 保存数据，不再更新

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)  # 线性层，特征维度变为原来的 3倍
        self.attn_drop = nn.Dropout(attn_drop)  # 随机丢弃神经元，丢弃率默认为 0.0
        self.proj = nn.Linear(dim, dim)  # 线性层，特征维度不变

        self.proj_drop = nn.Dropout(proj_drop)  # 随机丢弃神经元，丢弃率默认为 0.0

        trunc_normal_(self.relative_position_bias_table, std=.02)  # 截断正态分布，限制标准差为 0.02
        self.softmax = nn.Softmax(dim=-1)  # 激活函数 softmax

        # 定义前向传播

    def forward(self, x, mask=None):
        """
        输入:
            x: 输入特征图，结构为 [num_windows*B, N, C]
            mask: (0/-inf) mask, 结构为 [num_windows, Wh*Ww, Wh*Ww] 或者没有 mask
        """
        B_, N, C = x.shape  # 输入特征图的结构
        # 将特征图的通道维度按照注意力头的个数重新划分，并再做交换维度操作
        qkv = self.qkv(x).reshape(B_, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4).contiguous()
        q, k, v = qkv[0], qkv[1], qkv[2]  # 方便后续写代码，重新赋值

        # q 乘以缩放因子
        q = q * self.scale
        # @ 代表常规意义上的矩阵相乘
        attn = (q @ k.transpose(-2, -1))  # q 和 k 相乘后并交换最后两个维度

        # 相对位置偏移，结构为 [Wh*Ww, Wh*Ww, num_heads]
        relative_position_bias = self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
            self.window_size[0] * self.window_size[1], self.window_size[0] * self.window_size[1], -1)
        # 相对位置偏移交换维度，结构为 [num_heads, Wh*Ww, Wh*Ww]
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()
        attn = attn + relative_position_bias.unsqueeze(0)  # 带相对位置偏移的注意力图

        if mask is not None:  # 判断是否有 mask
            nW = mask.shape[0]  # mask 的宽
            # 注意力图与 mask 相加
            attn = attn.view(B_ // nW, nW, self.num_heads, N, N) + mask.unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N, N)  # 恢复注意力图原来的结构
            attn = self.softmax(attn)  # 激活注意力图 [0, 1] 之间
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)  # 随机设置注意力图中的部分值为 0
        # 注意力图与 v 相乘得到新的注意力图
        x = (attn @ v).transpose(1, 2).reshape(B_, N, C)
        x = self.proj(x)  # 通过线性层
        x = self.proj_drop(x)  # 随机设置新注意力图中的部分值为 0
        return x

# 一步实现既上采样也降维
class UpsampleOneStep(nn.Sequential):
    """一步上采样与前边上采样模块不同之处在于该模块只有一个卷积层和一个 pixelshuffle 层

    输入:
        scale (int): 缩放因子，支持 2^n and 3.
        num_feat (int): 中间特征的通道数.
    """

    def __init__(self, scale, num_feat, num_out_ch, input_resolution=None):
        self.num_feat = num_feat  # 中间特征的通道数
        self.input_resolution = input_resolution  # 输入分辨率
        m = []
        m.append(nn.Conv2d(num_feat, (scale ** 2) * num_out_ch, 3, 1, 1))  # 卷积层
        m.append(nn.PixelShuffle(scale))  # pixelshuffle 上采样 scale 倍
        super(UpsampleOneStep, self).__init__(*m)


class PatchEmbed(nn.Module):
    def __init__(self,img_size = 32,patch_size = 4,in_chans = 5, embed_dim = 96, norm_layer=None):
        super().__init__()
        patches_resolution = [img_size // patch_size,img_size // patch_size]
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_channel = in_chans
        self.embed_dim = embed_dim

        if norm_layer is not None:
            self.norm_layer = norm_layer(embed_dim)
        else:
            self.norm_layer = None

    def forward(self,x):
        x = x.flatten(2).transpose(1,2)
        if self.norm_layer is not None:
            x = self.norm_layer(x)
        return x

class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x



class DSSTnet(nn.Module):
    def __init__(self, opt):
        super(DSSTnet, self).__init__()
        self.img_range = opt.img_range
        self.num_in_chan = opt.in_chans
        self.num_out_chan = opt.out_chans
        self.num_feats = opt.num_feats
        self.upscale = opt.upscale_factor
        self.embed_dim = opt.embed_dim
        self.img_size = opt.img_size
        self.patch_size = opt.patch_size
        self.depths = opt.depths
        self.num_layers = len(self.depths)
        self.num_heads = opt.num_heads
        self.window_size = opt.window_size
        self.mlp_ration = opt.mlp_ratio
        self.qkv_bias = opt.qkv_bias
        self.qk_scale = opt.qk_scale
        self.drop_rate = opt.drop_rate
        self.attn_drop_rate = opt.attn_drop_rate
        self.drop_path_rate = opt.drop_path_rate
        self.norm_layer = nn.LayerNorm
        self.embed_layer = opt.embed_dim
        self.num_features = opt.embed_dim
        self.ape = False
        self.patch_norm = True
        self.use_checkpoint = False
        self.resi_connection = '1conv'


        self.conv_first = nn.Conv2d(self.num_in_chan,self.embed_dim,3,1,1)

        # 将图像分割成多个不重叠的patch
        self.patch_embed = PatchEmbed(
            img_size=self.img_size, patch_size=self.patch_size, in_chans=self.embed_dim, embed_dim=self.embed_dim,
            norm_layer=self.norm_layer if self.patch_norm else None
        )
        num_patches = self.patch_embed.num_patches  # 分割得到patch的个数
        patches_resolution = self.patch_embed.patches_resolution    # 分割得到patch的分辨率
        self.patch_resolution = patches_resolution

        # 将多个不重叠的patch合并成图像
        self.patch_unembed = PatchUnEmbed(
            img_size=self.img_size, patch_size=self.patch_size, in_chans=self.embed_dim, embed_dim=self.embed_dim,
            norm_layer=self.norm_layer if self.patch_norm else None
        )

        if self.ape:
            self.absolute_pos_embed = nn.Parameter(torch.zeros(1, num_patches, self.embed_dim))
            trunc_normal_(self.absolute_pos_embed, std = .02)

        self.pos_drop = nn.Dropout(p = self.drop_rate)

        # 随机深度衰减规律，默认为 [0, 0.1] 进行24等分后的列表
        dpr = [x.item() for x in torch.linspace(0, self.drop_path_rate, sum(self.depths))]

        # Resdual Swin Transformer blocks(RSTB)
        # 残差Swin Transformer块
        self.layers = nn.ModuleList()
        for i_layer in range(self.num_layers):
            layer = RSTB(dim = self.embed_dim,
                          input_resolution=(patches_resolution[0],
                                            patches_resolution[1]),
                          depth = self.depths[i_layer],
                          num_heads= self.num_heads[i_layer],
                          window_size = self.window_size,
                          mlp_ratio = self.mlp_ration,
                          qkv_bias = self.qkv_bias, qk_scale= self.qk_scale,
                          drop = self.drop_rate, attn_drop= self.attn_drop_rate,
                          drop_path=dpr[sum(self.depths[:i_layer]):sum(self.depths[:i_layer + 1])],
                          norm_layer= self.norm_layer,
                          downsample=None,
                          use_checkpoint=self.use_checkpoint,
                          img_size= self.img_size,
                          patch_size= self.patch_size,
                          resi_connection= self.resi_connection)
            self.layers.append(layer)
        self.norm = self.norm_layer(self.num_features)

        self.conv_after_body = nn.Conv2d(self.embed_dim, self.embed_dim, 3, 1, 1)

        self.conv_before_upsample = nn.Sequential(nn.Conv2d(self.embed_dim, self.num_feats, 3, 1, 1),
                                                  nn.LeakyReLU(inplace=True))
        self.upsample = Upsample(self.upscale, self.num_feats)
        self.conv_last = nn.Conv2d(self.num_feats, self.num_out_chan, 3, 1, 1)
        self.shallow = shallow_feature()
        self.DWT_3D = DWT_3D(in_channels=2, trainable=True)
        self.IDWT_3D = IDWT_3D(in_channels=2, trainable=True)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):  # 判断是否为线性 Linear 层
            trunc_normal_(m.weight, std=.02)  # 截断正态分布，限制标准差为 0.02
            if m.bias is not None:  # 如果设置了偏置
                nn.init.constant_(m.bias, 0)  # 初始化偏置为 0
        elif isinstance(m, nn.LayerNorm):  # 判断是否为归一化 LayerNorm 层
            nn.init.constant_(m.bias, 0)  # 初始化偏置为 0
            nn.init.constant_(m.weight, 1.0)  # 初始化权重系数为 1

        # 检查图片(准确说是张量)的大小

    def check_image_size(self, x):
        _, _, h, w = x.size()  # 张量 x 的高和宽
        # h 维度要填充的个数
        mod_pad_h = (self.window_size - h % self.window_size) % self.window_size
        # w 维度要填充的个数
        mod_pad_w = (self.window_size - w % self.window_size) % self.window_size
        # 右填充 mod_pad_w 个值，下填充 mod_pad_h 个值，模式为反射(可以理解为以 x 的维度末尾为轴对折)
        x = F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')
        return x

        # 深层特征提取网络的前向传播

    def forward_features(self, x):
        x_size = (x.shape[2], x.shape[3])  # 张量 x 的高和宽
        x = self.patch_embed(x)  # 分割 x 为多个不重叠的 patch embeddings
        if self.ape:  # 绝对位置 embedding
            x = x + self.absolute_pos_embed  # x 加上对应的绝对位置 embedding
        x = self.pos_drop(x)  # 随机将x中的部分元素置 0

        for layer in self.layers:
            x = layer(x, x_size)  # x 通过多个串联的 RSTB

        x = self.norm(x)  # 对 RSTB 的输出进行归一化
        x = self.patch_unembed(x, x_size)  # 将多个不重叠的 patch 合并成图像

        return x

        # SWinIR 的前向传

    def forward(self, x, neigbor_l, neigbor_r, bicu, first_state=None):
        out = x.unsqueeze(1)
        left = neigbor_l
        right = neigbor_r


        left = self.check_image_size(left)
        right = self.check_image_size(right)
        out = self.check_image_size(out)  # 检查图片的大小，使高宽满足 window_size 的整数倍


        out = torch.cat((out, out), 1)
        out, left, right = out.unsqueeze(2), left.unsqueeze(2), right.unsqueeze(2)
        out = torch.cat((left, out, right), 2)
        LLL, LLH, LHL, LHH, HLL, HLH, HHL, HHH = self.DWT_3D(out)
        HLL, HLH, HHL, HHH = torch.zeros(LLL.shape), torch.zeros(LLL.shape), torch.zeros(LLL.shape), torch.zeros(
            LLL.shape)
        HLL, HLH, HHL, HHH = HLL.cuda(), HLH.cuda(), HHL.cuda(), HHH.cuda()
        out = self.IDWT_3D(LLL, LLH, LHL, LHH, HLL, HLH, HHL, HHH)
        left, out, right = out[:, :, 0, :, :], out[:, 1, 2, :, :], out[:, :, 3, :, :]
        out = out.unsqueeze(1)
        out = self.shallow(out, left, right)  # 输入卷积层
        out = (self.conv_after_body(self.forward_features(out)) + out).contiguous()  # 深度特征提取网络，引入残差
        out = self.conv_before_upsample(out)  # 上采样前进行卷积
        first_state = out

        out = self.conv_last(self.upsample(out))

        out = torch.add(out.squeeze(1), bicu.squeeze(1))  # 上采样后再通过输出卷积层

        out = out / self.img_range  # 最后的 x 除以灰度值范围再加上 RGB 均值

        return out, first_state  # 返回输出 x


